--[[
  GarajalInnervation
  Copyright 2012  -  Dratr @ Stormscale

  You are free to use this addon for non-commercial purposes as a user interface
  addon for the Blizzard Entertainment game World of Warcraft.
  
  If you incorporate all or part of this addon into another work, you may do so only for
  non-commercial purposes and only if attribution is given to the original author; such
  derived works shall not use this author's name or this addon's name for any purpose
  other than acknowledging the source of the original work (in particular such may
  not be used to promote the derived work in any way).
  
  This work may not be redistributed in any way without written consent of the author.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
  IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
  OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  OTHER DEALINGS IN THE SOFTWARE.
  
    
  ABOUT THE ADDON
  GarajalInnervation opens a frame during the Garajal encounter that displays information
  about the amount of the Spiritual Innervation buff applied to players.  This overcomes
  the limitation that Blizzard implemented this buff as a non-stacking one, and the only
  way in the stock UI to tell how much buff one has is to examine the tooltip text, which
  can be most cumbersome during a raid encounter!
  It can optionally show your own buff only (intended for DPS players) or that of the
  entire raid (intended for healers or raid leaders).
  It can also optionally announce to players via whisper or raid message when their
  buff has reached capacity and they are able to exit without missing out on further
  buff stacks.
  
  The messaging option only functions for the raid leader and only when not in LFR
  mode.
  
--]]


local GarajalInnervation = LibStub("AceAddon-3.0"):NewAddon("GarajalInnervation", "AceConsole-3.0","AceEvent-3.0")


local TESTING_NAME = nil

local AURA_NAME = "Spiritual Innervation"
local ROW_INSETS = 4
local BAR_HEIGHT_PAD = 3
local BAR_WIDTH = 80
local TEXT_WIDTH = 80
local FRAME_WIDTH = BAR_WIDTH + TEXT_WIDTH + 3 * ROW_INSETS

local titleOffset = 0;
local barHeight = 0;
local GetInstanceInfo = _G.GetInstanceInfo
local MyFrame = nil;


-- Make array of widgets, size = numraidmembers
--    Track # of active widgets
-- On changes...
--    create sorted array - power -> { unitnames array }
--    re-text widgets based on this array

-- unitname -> table with keys:
--     value -> numeric value of aura, 0 if doesn't have aura
--     exp -> exp-time of aura, undefine if no aura (not filled in yet)
--     name -> unitname
--     sid -> spell id of the aura
--     msgsent -> boolean, sent msg or not on max buff reached
local playerAuraVals = {};

-- Unit names font strings, array
local widgets = {};
-- Aura % font strings, array (parented to bars[])
local pctWidgets = {}
-- StatusBar frames for % bars, array
local bars = {}
-- Num entries in above arrays
local numWidgets = 0;
-- Num entries in above arrays actually used for current roster
local numRows = 0;
-- Num entries actually actively showing right now
local activeWidgets = 0;

-- Is aura processing enabled, the frame showing, etc?
-- Note: this is not the configuration option "enabled" - that's in the db
local isEnabled = false

-- When display mode is myself only, this contains our UnitName
-- Changed up only on roster rebuild
local selfOnlyName = nil

-- True when display mode is not myself only and at least one messaging type
-- is enabled
-- Changed up only on roster rebuild
local shouldSendMsg = false


function GarajalInnervation:OnInitialize()
    local defaults = {
        profile = {
            frameloc = {
                x = 0,
                y = 0,
                anchorFrom = "CENTER",
                anchorTo = "CENTER"
            },
            enabled = true,
            mode = "all"
        }
    }
    self.db = LibStub("AceDB-3.0"):New("GarajalInnervationDB", defaults)
    
    -- Options GUI setup
    local options = {
         handler = GarajalInnervation,
         type = "group",
         width = "full",
         args = {
             enable = {
                 name = "Enable",
                 width = "full",
                 desc = "Enable or disable GarajalInnervation",
                 type = "toggle",
                 set = function(info, val) 
                         self.db.profile.enabled = val
                         if (not val) then
                             self:OnDisable()
                         else
                             self:OnEnable()
                         end
                       end,
                 get = function(info) return self.db.profile.enabled end,
                 order = 10
             },
             toggleframe = {
                 name = "Toggle Frame Visibility",
                 desc = "Show/Hide frame outside of encounter to position on screen",
                 type = "execute",
                 func = "ToggleWindow"
             },
             mode = {
                 name = "Display Mode",
                 desc = "Controls which player's buffs are displayed.",
                 width = "double",
                 type = "select",
                 order = 20,
                 values = { 
                     all="Show everyone (for healers and raid leaders)", 
                     myself="Show only your buff (for DPS)"
                 },
                 style = "dropdown",
                 set = function(info, val) self.db.profile.mode = val end,
                 get = function(info) return self.db.profile.mode end
             },
             msgoptions = {
                 name = "Messaging on Buff Capacity",
                 desc = "Send this type of message when buff has reached capacity. Requires raid leader to function. Disabled in LFR.",
                 type = "multiselect",
                 disabled = function() return self.db.profile.mode ~= "all" end,
                 order = 21,
                 set = function(info, key, val) self.db.profile[info[#info]..key] = val end,
                 get = function(info, key) return self.db.profile[info[#info]..key] end,
                 values = { 
                     sendtell = "Send whisper to player",
                     sendraid = "Send raid message",
                     sendraidwarning = "Send raid warning"
                 }
             }
         }
    }
    LibStub("AceConfig-3.0"):RegisterOptionsTable("GarajalInnervation", 
                                    options, {"gvation"})
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions("GarajalInnervation")
    

    -- Create our frame only once, but don't show it yet
    MyFrame = CreateFrame("frame", "GarajalInnervation Frame");
    MyFrame:Hide()
    MyFrame:SetWidth(FRAME_WIDTH);
    MyFrame:SetHeight(200); -- Initial height - changed later
    MyFrame:SetMovable(true);
    MyFrame:EnableMouse(true);
--    MyFrame:RegisterForDrag("LeftButton")
    MyFrame:SetScript("OnMouseDown", MyFrame.StartMoving)
    MyFrame:SetScript("OnMouseUp", function(self, button)
        self:StopMovingOrSizing()
        local opts = GarajalInnervation.db.profile.frameloc
        local from, _, to, x, y = self:GetPoint()
        opts.anchorFrom = from
        opts.anchorTo = to
        opts.offsetx = x
        opts.offsety = y
    end)
    MyFrame:SetPoint(self.db.profile.frameloc.anchorFrom, UIParent,
                     self.db.profile.frameloc.anchorTo,
                     self.db.profile.frameloc.x,
                     self.db.profile.frameloc.y)
    MyFrame:SetFrameStrata("MEDIUM")
    MyFrame:SetBackdrop({
      --bgFile="Interface\\DialogFrame\\UI-DialogBox-Background", 
      tile=1, tileSize=32, edgeSize=32, 
      insets={left=0, right=0, 
              top=0, bottom=0}
    })
    local backdrop = MyFrame:CreateTexture()
    backdrop:SetTexture(0.11, 0.17, 0.67, 0.2)
    backdrop:SetAllPoints(MyFrame)
    
    local titlebg = MyFrame:CreateTexture()
    titlebg:SetWidth(FRAME_WIDTH)
    titlebg:SetTexture(0, 0, 0, 0.6)
    local title = MyFrame:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
    title:SetText("Spiritual Innervation")
    title:SetFont("Fonts\\FRIZQT__.TTF", 8)
    title:SetTextColor(0.9, 0.9, 0.9, 1)
    titleOffset = title:GetHeight() + 2 * 2
    titlebg:SetHeight(titleOffset)
    titlebg:SetPoint("TOPLEFT", MyFrame, "TOPLEFT")
    title:SetPoint("TOPLEFT", MyFrame, "TOPLEFT")
    title:SetJustifyH("CENTER")
    title:SetJustifyV("MIDDLE")
    title:SetHeight(titleOffset)
    title:SetWidth(FRAME_WIDTH)
    titleOffset = titleOffset + ROW_INSETS
end


function GarajalInnervation:OnEnable()	
    -- self:RegisterEvent("ZONE_CHANGED_NEW_AREA","ZoneChangedNewArea")
    if (not self.db.profile.enabled) then
        return
    end

    -- Reset last zone vars incase just enabled
    self.lastzone = nil

    self:RegisterEvent("UPDATE_INSTANCE_INFO","ZoneChangedNewArea")
    -- Read zone as we weren't tracking this
    self:ZoneChangedNewArea()
end


function GarajalInnervation:OnDisable()
    self:DisableBossCombat()
    self:UnregisterAllEvents()
end


-- called from Config
function GarajalInnervation:ToggleWindow()
    if (MyFrame:IsShown()) then
        MyFrame:Hide()
    else
        MyFrame:Show()
    end
end


function GarajalInnervation:CombatStart(event)
    if (not isEnabled and self:CheckZone() and self:CheckBoss()) then
        self:EnableBossCombat()
    end
end

function GarajalInnervation:CombatEnd(event)
    self:DisableBossCombat()
--    self:Print("Combat ended!");
end

function GarajalInnervation:DisableBossCombat()
    if (not isEnabled) then
        return
    end
    self:UnregisterEvent("UNIT_AURA")
    self:UnregisterEvent("GROUP_ROSTER_UPDATE")
    MyFrame:Hide()
    self:RemoveAllWidgets()
    isEnabled = false
end


function GarajalInnervation:EnableBossCombat()
    if (isEnabled) then
        return
    end
    self:RebuildFromRoster()
    MyFrame:Show()
    self:RegisterEvent("UNIT_AURA","UnitAura")
    self:RegisterEvent("GROUP_ROSTER_UPDATE", "RosterEvent")
    isEnabled = true
--    self:Print("Enabled!");
end

function GarajalInnervation:CheckZone()
    return self.lastzone == "Mogu'shan Vaults"
end

function GarajalInnervation:CheckBoss()
    local exists = UnitExists("boss1");
    local dead = UnitIsDeadOrGhost("boss1");
    local guid = UnitGUID("boss1");
    if (exists and not dead and guid ~= nil) then
         local cid = tonumber(guid:sub(6,10), 16)
         return cid == 60143
    end
    return false
end

function GarajalInnervation:ZoneChangedNewArea(event)
    local zone, isLFR = self:GetInstanceDetails()

    if zone == self.lastzone then
      -- Same zone, bail
      return
    end

    self.lastzone = zone
    self.isLFR = isLFR
    
    if (self:CheckZone()) then
        -- Moved into the correct zone - enable our event calls
        -- self:Print("Entered MSV - Register combat events")
        self:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT", "CombatStart")
        self:RegisterEvent("PLAYER_REGEN_DISABLED", "CombatStart")
        self:RegisterEvent("PLAYER_REGEN_ENABLED", "CombatEnd")
    else
        self:UnregisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")
        self:UnregisterEvent("PLAYER_REGEN_DISABLED")
        self:UnregisterEvent("PLAYER_REGEN_ENABLED")
        self:DisableBossCombat()
    end
end




function GarajalInnervation:UnitAura(event, unitid)
    if (not isEnabled) then
        return
    end
    
    local unit = GetUnitName(unitid, true)
    
    if (selfOnlyName and selfOnlyName ~= unit) then
        return
    end

    if (not playerAuraVals[unit]) then
        -- Bail if they aren't on the roster.  Could be a pet
        return
    end

    --local auraname = "Earthliving"
    
    local _, spell, spellid, val1, val2, val3
    spell, _, _, _, _, _, _, _, _, _, spellid, _, _, val1, val2, val3 = UnitBuff(unitid, AURA_NAME)
    -- Determine if there is any change in the aura value
    -- so we know if we need to gui update or not
    local oldVal = playerAuraVals[unit]["value"]
    local newVal = oldVal

    if (not spell) then
        -- Player does not have aura
        newVal = 0
    elseif (spellid == 117543) then
        -- Healer buff uses val1, units are mana/s regen
        newVal = val1
    else
        -- DPS buff uses val3, values are 0 -> 22% buff
        newVal = val3
    end

    if (oldVal ~= newVal) then
        playerAuraVals[unit]["value"] = newVal;
        playerAuraVals[unit]["sid"] = spellid;
        self:UpdateWidgets();
        if (shouldSendMsg) then
            -- Decide if it is time to send a message
            if (newVal == 0) then
                playerAuraVals[unit]["msgsent"] = false
            elseif (newVal > 21 and not playerAuraVals[unit]["msgsent"] and playerAuraVals[unit]["sid"] ~= 117543) then
                self:SendMessages(unit)
                playerAuraVals[unit]["msgsent"] = true
            end
        end
    end
    
        
end

function SendMessages(unitname)
    if (db.profile.msgoptionssendtell) then
        SendChatMessage("You have reached max for Spiritual Innervation!", "WHISPER", nil, unitname)
    end
    
    if (db.profile.msgoptionssendraid) then
        SendChatMessage(unitname .. " has reached max for Spiritual Innervation!", "RAID_WARNING")
    end
    
    if (db.profile.msgoptionssendraidwarning) then
        SendChatMessage(unitname .. " has reached max for Spiritual Innervation!", "RAID")
    end
end

function GarajalInnervation:UpdateWidgets()
    local valsToInfos = {}
    local e, n, inf, i

    for n,inf in pairs(playerAuraVals) do
        local val = inf["value"]
        if (val ~= 0) then
            e = valsToInfos[inf["value"]];
            if (not e) then
                e = {}
                valsToInfos[inf["value"]] = e
            end
            table.insert(e, inf)
        end
    end
    
    -- convert to numerical array - array elements are:
    --    array of idx, auraval
    local sortArray = {}
    for e in pairs(valsToInfos) do
        table.insert(sortArray, e)
    end
    
    -- Sort in reverse order
    table.sort(sortArray, function(a, b) return a>b end)
    
    local idx = 1
    for i, n in ipairs(sortArray) do
        local i2, inftab
        for i2, inftab in pairs(valsToInfos[n]) do
            local v = inftab["value"]
            if (inftab["sid"] == 117543) then
                -- healer
                bars[idx]:SetStatusBarColor(0.08, 0.18, 0.89, 0.8)
                -- Normalize over 100k mana per sec - seems like an ok number
                --    based on LFR testing
                v = v / 1000
                pctWidgets[idx]:SetText(string.format("%.0f%%", v))
            else
                bars[idx]:SetStatusBarColor(0.89, 0.8, 0.08, 0.8)
                local v2 = v
                v = v * 100 / 22
                pctWidgets[idx]:SetText(string.format("%.0f%% (%.0f%%)", v, v2))
            end
            widgets[idx]:SetText(string.format("%s", inftab["name"]))
            bars[idx]:SetValue(v)
            widgets[idx]:Show()
            bars[idx]:Show()
            idx = idx + 1
            if (idx > numRows) then
                break
            end
        end
    end
    
    for i = idx, activeWidgets do
        widgets[i]:Hide()
        bars[i]:Hide()
    end

    activeWidgets = idx - 1
end

function GarajalInnervation:RemoveAllWidgets()
    local idx
    for idx = 1,activeWidgets do
        widgets[idx]:Hide()
        bars[idx]:Hide()
    end
    activeWidgets = 0
end

function GarajalInnervation:CreateNewWidget()
    local playercount = numWidgets
    local widget = MyFrame:CreateFontString(nil, "OVERLAY")
    widget:SetFont("Fonts\\FRIZQT__.TTF", 8)
    widget:SetTextColor(0.9, 0.9, 0.9, 1)
    widget:SetText("--");
    widget:SetWidth(TEXT_WIDTH)
    widget:SetWordWrap(false)
    widget:SetJustifyH("LEFT")
    widget:Hide();
    local textHeight = widget:GetHeight();
    barHeight = BAR_HEIGHT_PAD * 2 + textHeight
    widget:SetPoint("TOPLEFT", MyFrame, "TOPLEFT", ROW_INSETS, -1 * (titleOffset + (ROW_INSETS + barHeight) * playercount + BAR_HEIGHT_PAD));
    local bar = CreateFrame("StatusBar", nil, MyFrame)
    bar:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar")
    bar:SetMinMaxValues(0, 100)
    bar:SetValue(48)
    bar:SetWidth(BAR_WIDTH)
    bar:SetHeight(barHeight)
    bar:SetPoint("TOPLEFT", MyFrame, "TOPLEFT", 2 * ROW_INSETS + TEXT_WIDTH, -1 * (titleOffset + (ROW_INSETS + barHeight) * playercount));
    bar:SetStatusBarColor(0.89, 0.8, 0.08, 0.8)
    local backdrop = bar:CreateTexture()
    backdrop:SetTexture(0, 0, 0, 1)
    backdrop:SetAllPoints(bar)
    local pctText = bar:CreateFontString(nil, "OVERLAY")
    pctText:SetFont("Fonts\\FRIZQT__.TTF", 8)
    pctText:SetText("0%")
    pctText:SetAllPoints(bar)
    pctText:SetTextColor(0.9, 0.9, 0.9, 1)
    bar:Hide();
    
    numWidgets = numWidgets + 1
    widgets[numWidgets] = widget;
    pctWidgets[numWidgets] = pctText;
    bars[numWidgets] = bar;
end


function GarajalInnervation:RosterEvent(event)
    self:RebuildFromRoster()
end

function GarajalInnervation:RebuildFromRoster()
    local playercount = 0
    local idx
    playerAuraVals = {}
    self:RemoveAllWidgets()
    
    if (self.db.profile.mode == "myself") then
        selfOnlyName = GetUnitName("player", true)
        shouldSendMsg = false
    else
        selfOnlyName = nil
        if (UnitIsGroupLeader("player") and not self.isLFR) then
            shouldSendMsg = db.profile.msgoptionssendtell or
                            db.profile.msgoptionssendraid or
                            db.profile.msgoptionssendraidwarning
        else
            shouldSendMsg = false
        end
    end

    for idx = 1,MAX_RAID_MEMBERS do
        local name
        if (TESTING_NAME and idx == 1) then
            name = TESTING_NAME
        else
            name = GetRaidRosterInfo(idx)
        end
        if (name) then
            name = GetUnitName(name, true);
            
            if (not selfOnlyName or name == selfOnlyName) then

                local av = {}
                av["value"] = 0
                av["exp"] = 0
                av["name"] = name
                av["msgsent"] = false
                playercount = playercount + 1

                -- Cap total widgets at 10
                if (playercount > numWidgets and numWidgets < 10) then
                    self:CreateNewWidget()
                end
                
                
                
                playerAuraVals[name] = av
            end
        end
    end
    
    numRows = numWidgets
    if (playercount < numWidgets) then
        numRows = playercount
    end
    MyFrame:SetHeight(titleOffset + (ROW_INSETS + barHeight) * numRows);
end



-- returns zone, boolean-isLFR
function GarajalInnervation:GetInstanceDetails()
    local zone, type, diff = GetInstanceInfo()
    local islfr = diff == 7;
    
    return zone, islfr
end

